// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

package com.mapmyindia.mapmyindiagl;

import android.Manifest;
import android.annotation.SuppressLint;
import android.content.Context;
import android.content.pm.PackageManager;
import android.content.res.AssetFileDescriptor;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.PointF;
import android.graphics.RectF;
import android.location.Location;
import android.os.Build;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.Gravity;
import android.view.View;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.lifecycle.DefaultLifecycleObserver;
import androidx.lifecycle.Lifecycle;
import androidx.lifecycle.LifecycleOwner;

import com.google.gson.Gson;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsCallbcak;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsDetail;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsPlugin;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsRequest;
import com.mapmyindia.sdk.geoanalytics.MapmyIndiaGeoAnalyticsType;
import com.mapmyindia.sdk.geojson.Feature;
import com.mapmyindia.sdk.geojson.FeatureCollection;
import com.mapmyindia.sdk.maps.MapView;
import com.mapmyindia.sdk.maps.MapmyIndiaMap;
import com.mapmyindia.sdk.maps.MapmyIndiaMapOptions;
import com.mapmyindia.sdk.maps.OnMapReadyCallback;
import com.mapmyindia.sdk.maps.Style;
import com.mapmyindia.sdk.maps.camera.CameraPosition;
import com.mapmyindia.sdk.maps.camera.CameraUpdate;
import com.mapmyindia.sdk.maps.constants.MapmyIndiaConstants;
import com.mapmyindia.sdk.maps.geometry.LatLng;
import com.mapmyindia.sdk.maps.geometry.LatLngBounds;
import com.mapmyindia.sdk.maps.geometry.LatLngQuad;
import com.mapmyindia.sdk.maps.geometry.VisibleRegion;
import com.mapmyindia.sdk.maps.location.LocationComponent;
import com.mapmyindia.sdk.maps.location.LocationComponentOptions;
import com.mapmyindia.sdk.maps.location.OnCameraTrackingChangedListener;
import com.mapmyindia.sdk.maps.location.engine.LocationEngine;
import com.mapmyindia.sdk.maps.location.engine.LocationEngineCallback;
import com.mapmyindia.sdk.maps.location.engine.LocationEngineProvider;
import com.mapmyindia.sdk.maps.location.engine.LocationEngineResult;
import com.mapmyindia.sdk.maps.location.modes.CameraMode;
import com.mapmyindia.sdk.maps.location.modes.RenderMode;
import com.mapmyindia.sdk.maps.offline.OfflineManager;
import com.mapmyindia.sdk.maps.style.expressions.Expression;
import com.mapmyindia.sdk.maps.style.layers.CircleLayer;
import com.mapmyindia.sdk.maps.style.layers.FillLayer;
import com.mapmyindia.sdk.maps.style.layers.HillshadeLayer;
import com.mapmyindia.sdk.maps.style.layers.Layer;
import com.mapmyindia.sdk.maps.style.layers.LineLayer;
import com.mapmyindia.sdk.maps.style.layers.PropertyValue;
import com.mapmyindia.sdk.maps.style.layers.RasterLayer;
import com.mapmyindia.sdk.maps.style.layers.SymbolLayer;
import com.mapmyindia.sdk.maps.style.model.MapmyIndiaStyle;
import com.mapmyindia.sdk.maps.style.sources.GeoJsonSource;
import com.mapmyindia.sdk.maps.style.sources.ImageSource;
import com.mapmyindia.sdk.plugin.annotation.Annotation;
import com.mapmyindia.sdk.plugin.annotation.Circle;
import com.mapmyindia.sdk.plugin.annotation.CircleManager;
import com.mapmyindia.sdk.plugin.annotation.CircleOptions;
import com.mapmyindia.sdk.plugin.annotation.Fill;
import com.mapmyindia.sdk.plugin.annotation.FillManager;
import com.mapmyindia.sdk.plugin.annotation.FillOptions;
import com.mapmyindia.sdk.plugin.annotation.Line;
import com.mapmyindia.sdk.plugin.annotation.LineManager;
import com.mapmyindia.sdk.plugin.annotation.LineOptions;
import com.mapmyindia.sdk.plugin.annotation.OnAnnotationClickListener;
import com.mapmyindia.sdk.plugin.annotation.Symbol;
import com.mapmyindia.sdk.plugin.annotation.SymbolManager;
import com.mapmyindia.sdk.plugin.annotation.SymbolOptions;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import io.flutter.plugin.common.BinaryMessenger;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.platform.PlatformView;

/**
 * Controller of a single MapboxMaps MapView instance.
 */
@SuppressLint("MissingPermission")
final class MapboxMapController
        implements DefaultLifecycleObserver,
        MapmyIndiaMap.OnCameraIdleListener,
        MapmyIndiaMap.OnCameraMoveListener,
        MapmyIndiaMap.OnCameraMoveStartedListener,
        MapView.OnDidBecomeIdleListener,
        OnAnnotationClickListener,
        MapmyIndiaMap.OnMapClickListener,
        MapmyIndiaMap.OnMapLongClickListener,
        MapboxMapOptionsSink,
        MethodChannel.MethodCallHandler,
        OnMapReadyCallback,
        OnCameraTrackingChangedListener,
        OnSymbolTappedListener,
        OnLineTappedListener,
        OnCircleTappedListener,
        OnFillTappedListener,
        MapmyIndiaGeoAnalyticsCallbcak,
        PlatformView {
    private static final String TAG = "MapboxMapController";
    private final int id;
    private final MethodChannel methodChannel;
    private final MapboxMapsPlugin.LifecycleProvider lifecycleProvider;
    private MapView mapView;
    private MapmyIndiaMap mapmyIndiaMap;
    private final Map<String, SymbolController> symbols;
    private final Map<String, LineController> lines;
    private final Map<String, CircleController> circles;
    private final Map<String, FillController> fills;
    private SymbolManager symbolManager;
    private LineManager lineManager;
    private CircleManager circleManager;
    private FillManager fillManager;
    private boolean trackCameraPosition = false;
    private boolean myLocationEnabled = false;
    private int myLocationTrackingMode = 0;
    private int myLocationRenderMode = 0;
    private boolean disposed = false;
    private final float density;
    private MethodChannel.Result mapReadyResult;
    private final MethodChannel errorMethodChannel;
    private final Context context;
    private final String styleStringInitial;
    private final String styleMapmyIndiaStringInitial;
    private LocationComponent locationComponent = null;
    private LocationEngine locationEngine = null;
    private LocationEngineCallback<LocationEngineResult> locationEngineCallback = null;
    //  private LocalizationPlugin localizationPlugin;
    private MapmyIndiaGeoAnalyticsPlugin geoAnalyticsPlugin;
    private Style style;
    private List<String> annotationOrder;
    private List<String> annotationConsumeTapEvents;
    private Set<String> featureLayerIdentifiers;
    private LatLngBounds bounds = null;
    private static final String annotationManagerNotCreatedErrorCode = "NO ANNOTATION MANAGER";
    private static final String annotationManagerNotCreatedErrorMessage = "To use %ss please add it to the annotation list";
    private boolean showGeoAnalyticsInfoWindow = true;

    MapboxMapController(
            int id,
            Context context,
            BinaryMessenger messenger,
            MapboxMapsPlugin.LifecycleProvider lifecycleProvider,
            MapmyIndiaMapOptions options,
            String mapmyIndiaStyle,
            String styleStringInitial,
            List<String> annotationOrder,
            List<String> annotationConsumeTapEvents) {
        MapBoxUtils.getMapbox(context);
        this.id = id;
        this.context = context;
        this.styleStringInitial = styleStringInitial;
        this.styleMapmyIndiaStringInitial = mapmyIndiaStyle;
        this.mapView = new MapView(context, options);
        this.featureLayerIdentifiers = new HashSet<>();
        this.symbols = new HashMap<>();
        this.lines = new HashMap<>();
        this.circles = new HashMap<>();
        this.fills = new HashMap<>();
        this.density = context.getResources().getDisplayMetrics().density;
        this.lifecycleProvider = lifecycleProvider;
        methodChannel = new MethodChannel(messenger, "plugins.flutter.io/mapmyindia_maps_" + id);
        methodChannel.setMethodCallHandler(this);
        errorMethodChannel = new MethodChannel(messenger, "plugins.flutter.io/mapmyindia_maps_error_" + id);
        this.annotationOrder = annotationOrder;
        this.annotationConsumeTapEvents = annotationConsumeTapEvents;
    }

    @Override
    public View getView() {
        return mapView;
    }

    void init() {
        lifecycleProvider.getLifecycle().addObserver(this);
        mapView.getMapAsync(this);
    }

    private void moveCamera(CameraUpdate cameraUpdate) {
        mapmyIndiaMap.moveCamera(cameraUpdate);
    }

    private void animateCamera(CameraUpdate cameraUpdate) {
        mapmyIndiaMap.animateCamera(cameraUpdate);
    }

    private CameraPosition getCameraPosition() {
        return trackCameraPosition ? mapmyIndiaMap.getCameraPosition() : null;
    }

    private SymbolController symbol(String symbolId) {
        final SymbolController symbol = symbols.get(symbolId);
        if (symbol == null) {
            throw new IllegalArgumentException("Unknown symbol: " + symbolId);
        }
        return symbol;
    }

    private LineBuilder newLineBuilder() {
        return new LineBuilder(lineManager);
    }

    private void removeLine(String lineId) {
        final LineController lineController = lines.remove(lineId);
        if (lineController != null) {
            lineController.remove(lineManager);
        }
    }

    private LineController line(String lineId) {
        final LineController line = lines.get(lineId);
        if (line == null) {
            throw new IllegalArgumentException("Unknown line: " + lineId);
        }
        return line;
    }

    private CircleBuilder newCircleBuilder() {
        return new CircleBuilder(circleManager);
    }

    private void removeCircle(String circleId) {
        final CircleController circleController = circles.remove(circleId);
        if (circleController != null) {
            circleController.remove(circleManager);
        }
    }

    private CircleController circle(String circleId) {
        final CircleController circle = circles.get(circleId);
        if (circle == null) {
            throw new IllegalArgumentException("Unknown circle: " + circleId);
        }
        return circle;
    }

    private FillBuilder newFillBuilder() {
        return new FillBuilder(fillManager);
    }

    private void removeFill(String fillId) {
        final FillController fillController = fills.remove(fillId);
        if (fillController != null) {
            fillController.remove(fillManager);
        }
    }

    private FillController fill(String fillId) {
        final FillController fill = fills.get(fillId);
        if (fill == null) {
            throw new IllegalArgumentException("Unknown fill: " + fillId);
        }
        return fill;
    }

    @Override
    public void onMapReady(MapmyIndiaMap mapboxMap) {
        this.mapmyIndiaMap = mapboxMap;
        if (mapReadyResult != null) {
            mapReadyResult.success(null);
            mapReadyResult = null;
        }
        mapboxMap.addOnCameraMoveStartedListener(this);
        mapboxMap.addOnCameraMoveListener(this);
        mapboxMap.addOnCameraIdleListener(this);


        mapView.addOnStyleImageMissingListener((id) -> {
            DisplayMetrics displayMetrics = context.getResources().getDisplayMetrics();
            final Bitmap bitmap = getScaledImage(id, displayMetrics.density);
            if (bitmap != null) {
                mapboxMap.getStyle().addImage(id, bitmap);
            }
        });

        mapView.addOnDidBecomeIdleListener(this);

        if (styleMapmyIndiaStringInitial != null) {
            mapboxMap.setMapmyIndiaStyle(styleMapmyIndiaStringInitial, onStyleLoadedCallback);
        } else {
            mapboxMap.getStyle(new Style.OnStyleLoaded() {
                @Override
                public void onStyleLoaded(@NonNull Style style) {
                    onAfterStyleLoaded(style);
                }
            });
        }
//        setStyleString(styleStringInitial);
        // updateMyLocationEnabled();
    }

    @Override
    public void onMapError(int i, String s) {
        final Map<String, Object> map = new HashMap<>(2);
        map.put("errorCode", i);
        map.put("message", s);
        errorMethodChannel.invokeMethod("map#onMapError", map);
    }

    private void onAfterStyleLoaded(Style style) {
        MapboxMapController.this.style = style;

        // only add managers once to avoid issues with getLayerId after a style switch
        if (symbolManager == null && circleManager == null && lineManager == null && fillManager == null) {
            final List<String> orderReversed = new ArrayList<String>(annotationOrder);
            Collections.reverse(orderReversed);
            String belowLayer = null;
            for (String annotationType : orderReversed) {
                switch (annotationType) {
                    case "AnnotationType.fill":
                        belowLayer = enableFillManager(style, belowLayer);
                        break;
                    case "AnnotationType.line":
                        belowLayer = enableLineManager(style, belowLayer);
                        break;
                    case "AnnotationType.circle":
                        belowLayer = enableCircleManager(style, belowLayer);
                        break;
                    case "AnnotationType.symbol":
                        belowLayer = enableSymbolManager(style, belowLayer);
                        break;
                    default:
                        throw new IllegalArgumentException("Unknown annotation type: " + annotationType + ", must be either 'fill', 'line', 'circle' or 'symbol'");
                }
            }
        }

        if (myLocationEnabled) {
            enableLocationComponent(style);
        }

        if (null != bounds) {
            mapmyIndiaMap.setLatLngBoundsForCameraTarget(bounds);
        }

        // needs to be placed after SymbolManager#addClickListener,
        // is fixed with 0.6.0 of annotations plugin
        mapmyIndiaMap.addOnMapClickListener(MapboxMapController.this);
        mapmyIndiaMap.addOnMapLongClickListener(MapboxMapController.this);
        geoAnalyticsPlugin = new MapmyIndiaGeoAnalyticsPlugin(mapView, mapmyIndiaMap);
        geoAnalyticsPlugin.setGeoAnalyticsCallback(this);
        geoAnalyticsPlugin.shouldShowInfoWindow(showGeoAnalyticsInfoWindow);
//	    localizationPlugin = new LocalizationPlugin(mapView, mapboxMap, style);

        methodChannel.invokeMethod("map#onStyleLoaded", null);
    }

    @Override
    public void setStyleString(String styleString) {
        // Check if json, url, absolute path or asset path:
        if (styleString == null || styleString.isEmpty()) {
            Log.e(TAG, "setStyleString - string empty or null");
        } else if (styleString.startsWith("{") || styleString.startsWith("[")) {
            mapmyIndiaMap.setStyle(new Style.Builder().fromJson(styleString), onStyleLoadedCallback);
        } else if (styleString.startsWith("/")) {
            // Absolute path
//      mapboxMap.setStyle(new Style.Builder().fromUri("file://" + styleString), onStyleLoadedCallback);
        } else if (
                !styleString.startsWith("http://") &&
                        !styleString.startsWith("https://") &&
                        !styleString.startsWith("mapbox://")) {
            // We are assuming that the style will be loaded from an asset here.
//      String key = MapboxMapsPlugin.flutterAssets.getAssetFilePathByName(styleString);
//      mapboxMap.setStyle(new Style.Builder().fromUri("asset://" + key), onStyleLoadedCallback);
        } else {
//      mapboxMap.setStyle(new Style.Builder().fromUri(styleString), onStyleLoadedCallback);
        }
    }

    @Override
    public void setMapmyIndiaStyle(String styleString) {
        if (styleString != null) {
            if (lineManager != null) {
                lines.clear();
                lineManager.clearAll();
            }
            if (fillManager != null) {
                fills.clear();
                fillManager.clearAll();
            }
            if (circleManager != null) {
                circles.clear();
                circleManager.clearAll();
            }
            if (symbolManager != null) {
                symbols.clear();
                symbolManager.clearAll();
            }
            mapmyIndiaMap.setMapmyIndiaStyle(styleString, onStyleLoadedCallback);
        }
    }

    Style.OnStyleLoaded onStyleLoadedCallback = new Style.OnStyleLoaded() {
        @Override
        public void onStyleLoaded(@NonNull Style style) {
            onAfterStyleLoaded(style);
        }
    };

    @SuppressWarnings({"MissingPermission"})
    private void enableLocationComponent(@NonNull Style style) {
        if (hasLocationPermission()) {
            locationEngine = LocationEngineProvider.getBestLocationEngine(context);
            LocationComponentOptions locationComponentOptions = LocationComponentOptions.builder(context)
                    .trackingGesturesManagement(true)
                    .build();
            locationComponent = mapmyIndiaMap.getLocationComponent();
            locationComponent.activateLocationComponent(context, style, locationComponentOptions);
            locationComponent.setLocationComponentEnabled(true);
            // locationComponent.setRenderMode(RenderMode.COMPASS); // remove or keep default?
            locationComponent.setLocationEngine(locationEngine);
            locationComponent.setMaxAnimationFps(30);
            updateMyLocationTrackingMode();
            setMyLocationTrackingMode(this.myLocationTrackingMode);
            updateMyLocationRenderMode();
            setMyLocationRenderMode(this.myLocationRenderMode);
            locationComponent.addOnCameraTrackingChangedListener(this);
        } else {
            Log.e(TAG, "missing location permissions");
        }
    }

    private void onUserLocationUpdate(Location location) {
        if (location == null) {
            return;
        }

        final Map<String, Object> userLocation = new HashMap<>(6);
        userLocation.put("position", new double[]{location.getLatitude(), location.getLongitude()});
        userLocation.put("speed", location.getSpeed());
        userLocation.put("altitude", location.getAltitude());
        userLocation.put("bearing", location.getBearing());
        userLocation.put("horizontalAccuracy", location.getAccuracy());
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            userLocation.put("verticalAccuracy", (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) ? location.getVerticalAccuracyMeters() : null);
        }
        userLocation.put("timestamp", location.getTime());

        final Map<String, Object> arguments = new HashMap<>(1);
        arguments.put("userLocation", userLocation);
        methodChannel.invokeMethod("map#onUserLocationUpdated", arguments);
    }

    private void addGeoJsonSource(String sourceName, String source) {
        FeatureCollection featureCollection = FeatureCollection.fromJson(source);
        GeoJsonSource geoJsonSource = new GeoJsonSource(sourceName, featureCollection);

        style.addSource(geoJsonSource);
    }

    private void setGeoJsonSource(String sourceName, String source) {
        FeatureCollection featureCollection = FeatureCollection.fromJson(source);
        GeoJsonSource geoJsonSource = style.getSourceAs(sourceName);

        geoJsonSource.setGeoJson(featureCollection);
    }

    private void addSymbolLayer(String layerName,
                                String sourceName,
                                String belowLayerId,
                                String sourceLayer,
                                PropertyValue[] properties,
                                Expression filter) {
        SymbolLayer symbolLayer = new SymbolLayer(layerName, sourceName);
        symbolLayer.setProperties(properties);
        if (sourceLayer != null) {
            symbolLayer.setSourceLayer(sourceLayer);
        }

        if (belowLayerId != null) {
            style.addLayerBelow(symbolLayer, belowLayerId);
        } else {
            style.addLayer(symbolLayer);
        }
        featureLayerIdentifiers.add(layerName);
    }

    private void addLineLayer(String layerName,
                              String sourceName,
                              String belowLayerId,
                              String sourceLayer,
                              PropertyValue[] properties,
                              Expression filter) {
        LineLayer lineLayer = new LineLayer(layerName, sourceName);
        lineLayer.setProperties(properties);
        if (sourceLayer != null) {
            lineLayer.setSourceLayer(sourceLayer);
        }

        if (belowLayerId != null) {
            style.addLayerBelow(lineLayer, belowLayerId);
        } else {
            style.addLayer(lineLayer);
        }
        featureLayerIdentifiers.add(layerName);
    }

    private void addFillLayer(String layerName,
                              String sourceName,
                              String belowLayerId,
                              String sourceLayer,
                              PropertyValue[] properties,
                              Expression filter) {
        FillLayer fillLayer = new FillLayer(layerName, sourceName);
        fillLayer.setProperties(properties);
        if (sourceLayer != null) {
            fillLayer.setSourceLayer(sourceLayer);
        }

        if (belowLayerId != null) {
            style.addLayerBelow(fillLayer, belowLayerId);
        } else {
            style.addLayer(fillLayer);
        }
        featureLayerIdentifiers.add(layerName);
    }

    private void addCircleLayer(String layerName,
                                String sourceName,
                                String belowLayerId,
                                String sourceLayer,
                                PropertyValue[] properties,
                                Expression filter) {
        CircleLayer circleLayer = new CircleLayer(layerName, sourceName);
        circleLayer.setProperties(properties);
        if (sourceLayer != null) {
            circleLayer.setSourceLayer(sourceLayer);
        }

        featureLayerIdentifiers.add(layerName);
        if (belowLayerId != null) {
            style.addLayerBelow(circleLayer, belowLayerId);
        } else {
            style.addLayer(circleLayer);
        }
    }

    private void addRasterLayer(String layerName,
                                String sourceName,
                                String belowLayerId,
                                PropertyValue[] properties,
                                Expression filter) {
        RasterLayer layer = new RasterLayer(layerName, sourceName);
        layer.setProperties(properties);

        if (belowLayerId != null) {
            style.addLayerBelow(layer, belowLayerId);
        } else {
            style.addLayer(layer);
        }
    }

    private void addHillshadeLayer(String layerName,
                                   String sourceName,
                                   String belowLayerId,
                                   PropertyValue[] properties,
                                   Expression filter) {
        HillshadeLayer layer = new HillshadeLayer(layerName, sourceName);
        layer.setProperties(properties);

        if (belowLayerId != null) {
            style.addLayerBelow(layer, belowLayerId);
        } else {
            style.addLayer(layer);
        }
    }

    private String enableSymbolManager(@NonNull Style style, @Nullable String belowLayer) {
        if (symbolManager == null) {
            symbolManager = new SymbolManager(mapView, mapmyIndiaMap, style, belowLayer);
            symbolManager.setIconAllowOverlap(true);
            symbolManager.setIconIgnorePlacement(true);
            symbolManager.setTextAllowOverlap(true);
            symbolManager.setTextIgnorePlacement(true);
            symbolManager.addClickListener(MapboxMapController.this::onAnnotationClick);
        }
        return symbolManager.getLayerId();
    }

    private String enableLineManager(@NonNull Style style, @Nullable String belowLayer) {
        if (lineManager == null) {
            lineManager = new LineManager(mapView, mapmyIndiaMap, style, belowLayer);
            lineManager.addClickListener(MapboxMapController.this::onAnnotationClick);
        }
        return lineManager.getLayerId();
    }

    private String enableCircleManager(@NonNull Style style, @Nullable String belowLayer) {
        if (circleManager == null) {
            circleManager = new CircleManager(mapView, mapmyIndiaMap, style, belowLayer);
            circleManager.addClickListener(MapboxMapController.this::onAnnotationClick);
        }
        return circleManager.getLayerId();
    }

    private String enableFillManager(@NonNull Style style, @Nullable String belowLayer) {
        if (fillManager == null) {
            fillManager = new FillManager(mapView, mapmyIndiaMap, style, belowLayer);
            fillManager.addClickListener(MapboxMapController.this::onAnnotationClick);
        }
        return fillManager.getLayerId();
    }

    private Feature firstFeatureOnLayers(RectF in) {
        if (style != null) {
            final List<Layer> layers = style.getLayers();
            final List<String> layersInOrder = new ArrayList<String>();
            for (Layer layer : layers) {
                String id = layer.getId();
                if (featureLayerIdentifiers.contains(id))
                    layersInOrder.add(id);
            }
            Collections.reverse(layersInOrder);

            for (String id : layersInOrder) {
                List<Feature> features = mapmyIndiaMap.queryRenderedFeatures(in, id);
                if (!features.isEmpty()) {
                    return features.get(0);
                }
            }
        }
        return null;
    }

    @Override
    public void onMethodCall(MethodCall call, MethodChannel.Result result) {

        switch (call.method) {
            case "map#waitForMap":
                if (mapmyIndiaMap != null) {
                    result.success(null);
                    return;
                }
                mapReadyResult = result;
                break;
            case "map#update": {
                Convert.interpretMapboxMapOptions(call.argument("options"), this, context);
                result.success(Convert.toJson(getCameraPosition()));
                break;
            }
            case "map#updateMyLocationTrackingMode": {
                int myLocationTrackingMode = call.argument("mode");
                setMyLocationTrackingMode(myLocationTrackingMode);
                result.success(null);
                break;
            }
            case "map#matchMapLanguageWithDeviceDefault": {
                try {
//		      localizationPlugin.matchMapLanguageWithDeviceDefault();
                    result.success(null);
                } catch (RuntimeException exception) {
                    Log.d(TAG, exception.toString());
                    result.error("MAPBOX LOCALIZATION PLUGIN ERROR", exception.toString(), null);
                }
                break;
            }
            case "map#getMapmyIndiaStyle": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                result.success(style.getMapmyIndiaStyle());
                break;
            }
            case "map#setMapLanguage": {
                final String language = call.argument("language");
                try {
//		      localizationPlugin.setMapLanguage(language);
                    result.success(null);
                } catch (RuntimeException exception) {
                    Log.d(TAG, exception.toString());
                    result.error("MAPBOX LOCALIZATION PLUGIN ERROR", exception.toString(), null);
                }
                break;
            }
            case "map#getVisibleRegion": {
                Map<String, Object> reply = new HashMap<>();
                VisibleRegion visibleRegion = mapmyIndiaMap.getProjection().getVisibleRegion();
                reply.put("sw", Arrays.asList(visibleRegion.nearLeft.getLatitude(), visibleRegion.nearLeft.getLongitude()));
                reply.put("ne", Arrays.asList(visibleRegion.farRight.getLatitude(), visibleRegion.farRight.getLongitude()));
                result.success(reply);
                break;
            }
            case "map#toScreenLocation": {
                Map<String, Object> reply = new HashMap<>();
                PointF pointf = mapmyIndiaMap.getProjection().toScreenLocation(new LatLng(call.argument("latitude"), call.argument("longitude")));
                reply.put("x", pointf.x);
                reply.put("y", pointf.y);
                result.success(reply);
                break;
            }
            case "map#toScreenLocationBatch": {
                double[] param = (double[]) call.argument("coordinates");
                double[] reply = new double[param.length];

                for (int i = 0; i < param.length; i += 2) {
                    PointF pointf = mapmyIndiaMap.getProjection().toScreenLocation(new LatLng(param[i], param[i + 1]));
                    reply[i] = pointf.x;
                    reply[i + 1] = pointf.y;
                }

                result.success(reply);
                break;
            }
            case "map#toLatLng": {
                Map<String, Object> reply = new HashMap<>();
                LatLng latlng = mapmyIndiaMap.getProjection().fromScreenLocation(new PointF(((Double) call.argument("x")).floatValue(), ((Double) call.argument("y")).floatValue()));
                reply.put("latitude", latlng.getLatitude());
                reply.put("longitude", latlng.getLongitude());
                result.success(reply);
                break;
            }
            case "map#getMetersPerPixelAtLatitude": {
                Map<String, Object> reply = new HashMap<>();
                Double retVal = mapmyIndiaMap.getProjection().getMetersPerPixelAtLatitude((Double) call.argument("latitude"));
                reply.put("metersperpixel", retVal);
                result.success(reply);
                break;
            }
            case "camera#move": {
                final CameraUpdate cameraUpdate = Convert.toCameraUpdate(call.argument("cameraUpdate"), mapmyIndiaMap, density);
                if (cameraUpdate != null) {
                    // camera transformation not handled yet
                    mapmyIndiaMap.moveCamera(cameraUpdate, new OnCameraMoveFinishedListener() {
                        @Override
                        public void onFinish() {
                            super.onFinish();
                            result.success(true);
                        }

                        @Override
                        public void onCancel() {
                            super.onCancel();
                            result.success(false);
                        }
                    });

                    // moveCamera(cameraUpdate);
                } else {
                    result.success(false);
                }
                break;
            }
            case "camera#animate": {
                final CameraUpdate cameraUpdate = Convert.toCameraUpdate(call.argument("cameraUpdate"), mapmyIndiaMap, density);
                final Integer duration = call.argument("duration");

                final OnCameraMoveFinishedListener onCameraMoveFinishedListener = new OnCameraMoveFinishedListener() {
                    @Override
                    public void onFinish() {
                        super.onFinish();
                        result.success(true);
                    }

                    @Override
                    public void onCancel() {
                        super.onCancel();
                        result.success(false);
                    }
                };
                if (cameraUpdate != null && duration != null) {
                    // camera transformation not handled yet
                    mapmyIndiaMap.animateCamera(cameraUpdate, duration, onCameraMoveFinishedListener);
                } else if (cameraUpdate != null) {
                    // camera transformation not handled yet
                    mapmyIndiaMap.animateCamera(cameraUpdate, onCameraMoveFinishedListener);
                } else {
                    result.success(false);
                }
                break;
            }
            case "camera#ease": {
                final CameraUpdate cameraUpdate = Convert.toCameraUpdate(call.argument("cameraUpdate"), mapmyIndiaMap, density);
                final Integer duration = call.argument("duration");

                final OnCameraMoveFinishedListener onCameraMoveFinishedListener = new OnCameraMoveFinishedListener() {
                    @Override
                    public void onFinish() {
                        super.onFinish();
                        result.success(true);
                    }

                    @Override
                    public void onCancel() {
                        super.onCancel();
                        result.success(false);
                    }
                };
                if (cameraUpdate != null && duration != null) {
                    // camera transformation not handled yet
                    mapmyIndiaMap.easeCamera(cameraUpdate, duration, onCameraMoveFinishedListener);
                } else if (cameraUpdate != null) {
                    // camera transformation not handled yet
                    mapmyIndiaMap.easeCamera(cameraUpdate);
                } else {
                    result.success(false);
                }
                break;
            }
            case "camera#movewithELoc": {
                OnCameraMoveFinishedListener onCameraMoveFinishedListener = new OnCameraMoveFinishedListener() {
                    @Override
                    public void onFinish() {
                        super.onFinish();
                        result.success(true);
                    }

                    @Override
                    public void onCancel() {
                        super.onCancel();
                        result.success(false);
                    }
                };
                Convert.moveCameraWithELoc(call.argument("cameraELocUpdate"), mapmyIndiaMap, onCameraMoveFinishedListener, density);
                break;
            }
            case "camera#animateWithELoc": {
                final Integer duration = call.argument("duration");
                final OnCameraMoveFinishedListener onCameraMoveFinishedListener = new OnCameraMoveFinishedListener() {
                    @Override
                    public void onFinish() {
                        super.onFinish();
                        result.success(true);
                    }

                    @Override
                    public void onCancel() {
                        super.onCancel();
                        result.success(false);
                    }
                };
                Convert.animateCameraWithELoc(call.argument("cameraELocUpdate"), mapmyIndiaMap, duration, onCameraMoveFinishedListener, density);
                break;
            }
            case "camera#easeWithELoc": {
                final Integer duration = call.argument("duration");
                final OnCameraMoveFinishedListener onCameraMoveFinishedListener = new OnCameraMoveFinishedListener() {
                    @Override
                    public void onFinish() {
                        super.onFinish();
                        result.success(true);
                    }

                    @Override
                    public void onCancel() {
                        super.onCancel();
                        result.success(false);
                    }
                };
                Convert.easeCameraWithELoc(call.argument("cameraELocUpdate"), mapmyIndiaMap, duration, onCameraMoveFinishedListener, density);
                break;
            }
            case "map#queryRenderedFeatures": {
                Map<String, Object> reply = new HashMap<>();
                List<Feature> features;

                String[] layerIds = ((List<String>) call.argument("layerIds")).toArray(new String[0]);

                List<Object> filter = call.argument("filter");
                JsonElement jsonElement = filter == null ? null : new Gson().toJsonTree(filter);
                JsonArray jsonArray = null;
                if (jsonElement != null && jsonElement.isJsonArray()) {
                    jsonArray = jsonElement.getAsJsonArray();
                }
                Expression filterExpression = jsonArray == null ? null : Expression.Converter.convert(jsonArray);
                if (call.hasArgument("x")) {
                    Double x = call.argument("x");
                    Double y = call.argument("y");
                    PointF pixel = new PointF(x.floatValue(), y.floatValue());
                    features = mapmyIndiaMap.queryRenderedFeatures(pixel, filterExpression, layerIds);
                } else {
                    Double left = call.argument("left");
                    Double top = call.argument("top");
                    Double right = call.argument("right");
                    Double bottom = call.argument("bottom");
                    RectF rectF = new RectF(left.floatValue(), top.floatValue(), right.floatValue(), bottom.floatValue());
                    features = mapmyIndiaMap.queryRenderedFeatures(rectF, filterExpression, layerIds);
                }
                List<String> featuresJson = new ArrayList<>();
                for (Feature feature : features) {
                    featuresJson.add(feature.toJson());
                }
                reply.put("features", featuresJson);
                result.success(reply);
                break;
            }
            case "map#setTelemetryEnabled": {
                final boolean enabled = call.argument("enabled");
//        MapmyIndia.getTelemetry().setUserTelemetryRequestState(enabled);
                result.success(null);
                break;
            }
            case "map#getTelemetryEnabled": {
//        final TelemetryEnabler.State telemetryState = TelemetryEnabler.retrieveTelemetryStateFromPreferences();
                result.success(false);
//        result.success(telemetryState == TelemetryEnabler.State.ENABLED);
                break;
            }
            case "map#invalidateAmbientCache": {
                OfflineManager fileSource = OfflineManager.getInstance(context);

                fileSource.invalidateAmbientCache(new OfflineManager.FileSourceCallback() {
                    @Override
                    public void onSuccess() {
                        result.success(null);
                    }

                    @Override
                    public void onError(@NonNull String message) {
                        result.error("MAPBOX CACHE ERROR", message, null);
                    }
                });
                break;
            }
            case "map#getMapmyIndiaAvailableStyles": {
                List<MapmyIndiaStyle> mapmyIndiaStyles = mapmyIndiaMap.getMapmyIndiaAvailableStyles();
                List<Map<String, Object>> list = new ArrayList<>();
                for (MapmyIndiaStyle style : mapmyIndiaStyles) {
                    Map<String, Object> map = new HashMap<>();
                    map.put("description", style.getDescription());
                    map.put("displayName", style.getDisplayName());
                    map.put("imageUrl", style.getImageUrl());
                    map.put("name", style.getName());
                    list.add(map);
                }
                result.success(list);
                break;
            }
            case "symbols#addAll": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                List<String> newSymbolIds = new ArrayList<String>();
                final List<Object> options = call.argument("options");
                List<SymbolOptions> symbolOptionsList = new ArrayList<SymbolOptions>();
                if (options != null) {
                    SymbolBuilder symbolBuilder;
                    for (Object o : options) {
                        symbolBuilder = new SymbolBuilder();
                        Convert.interpretSymbolOptions(o, symbolBuilder);
                        symbolOptionsList.add(symbolBuilder.getSymbolOptions());
                    }
                    if (!symbolOptionsList.isEmpty()) {
                        List<Symbol> newSymbols = symbolManager.create(symbolOptionsList);
                        String symbolId;
                        for (Symbol symbol : newSymbols) {
                            symbolId = String.valueOf(symbol.getId());
                            newSymbolIds.add(symbolId);
                            symbols.put(symbolId, new SymbolController(symbol, annotationConsumeTapEvents.contains("AnnotationType.symbol"), this));
                        }
                    }
                }
                result.success(newSymbolIds);
                break;
            }
            case "symbols#addJson": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                List<Map<String, Object>> newSymbolIds = new ArrayList<Map<String, Object>>();
                final Object json = call.argument("json");
                if (json != null) {
                    List<Symbol> newSymbols = symbolManager.create(Convert.toString(json));
                    String symbolId;
                    for (Symbol symbol : newSymbols) {
                        symbolId = String.valueOf(symbol.getId());
                        newSymbolIds.add(Convert.convertSymbolOptions(symbolId, symbol));
                        symbols.put(symbolId, new SymbolController(symbol, annotationConsumeTapEvents.contains("AnnotationType.symbol"), this));
                    }
                }
                result.success(newSymbolIds);
                break;
            }
            case "symbols#removeAll": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final ArrayList<String> symbolIds = call.argument("ids");
                SymbolController symbolController;

                List<Symbol> symbolList = new ArrayList<Symbol>();
                for (String symbolId : symbolIds) {
                    symbolController = symbols.remove(symbolId);
                    if (symbolController != null) {
                        symbolList.add(symbolController.getSymbol());
                    }
                }
                if (!symbolList.isEmpty()) {
                    symbolManager.clear(symbolList);
                }
                result.success(null);
                break;
            }
            case "symbol#update": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final String symbolId = call.argument("symbol");
                final SymbolController symbol = symbol(symbolId);
                Convert.interpretSymbolOptions(call.argument("options"), symbol);
                symbol.update(symbolManager);
                result.success(null);
                break;
            }
            case "symbol#getGeometry": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final String symbolId = call.argument("symbol");
                final SymbolController symbol = symbol(symbolId);
                final LatLng symbolLatLng = symbol.getGeometry();
                Map<String, Double> hashMapLatLng = new HashMap<>();
                hashMapLatLng.put("latitude", symbolLatLng.getLatitude());
                hashMapLatLng.put("longitude", symbolLatLng.getLongitude());
                result.success(hashMapLatLng);
            }
            case "symbolManager#iconAllowOverlap": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final Boolean value = call.argument("iconAllowOverlap");
                symbolManager.setIconAllowOverlap(value);
                result.success(null);
                break;
            }
            case "symbolManager#iconIgnorePlacement": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final Boolean value = call.argument("iconIgnorePlacement");
                symbolManager.setIconIgnorePlacement(value);
                result.success(null);
                break;
            }
            case "symbolManager#textAllowOverlap": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final Boolean value = call.argument("textAllowOverlap");
                symbolManager.setTextAllowOverlap(value);
                result.success(null);
                break;
            }
            case "symbolManager#textIgnorePlacement": {
                if (symbolManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "symbol"), null);
                    return;
                }
                final Boolean iconAllowOverlap = call.argument("textIgnorePlacement");
                symbolManager.setTextIgnorePlacement(iconAllowOverlap);
                result.success(null);
                break;
            }
            case "line#add": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                final LineBuilder lineBuilder = newLineBuilder();
                Convert.interpretLineOptions(call.argument("options"), lineBuilder);
                final Line line = lineBuilder.build();
                final String lineId = String.valueOf(line.getId());
                lines.put(lineId, new LineController(line, annotationConsumeTapEvents.contains("AnnotationType.line"), this));
                result.success(lineId);
                break;
            }
            case "line#remove": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                final String lineId = call.argument("line");
                removeLine(lineId);
                result.success(null);
                break;
            }
            case "line#addAll": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                List<String> newIds = new ArrayList<String>();
                final List<Object> options = call.argument("options");
                List<LineOptions> optionList = new ArrayList<LineOptions>();
                if (options != null) {
                    LineBuilder builder;
                    for (Object o : options) {
                        builder = newLineBuilder();
                        Convert.interpretLineOptions(o, builder);
                        optionList.add(builder.getLineOptions());
                    }
                    if (!optionList.isEmpty()) {
                        List<Line> newLines = lineManager.create(optionList);
                        String id;
                        for (Line line : newLines) {
                            id = String.valueOf(line.getId());
                            newIds.add(id);
                            lines.put(id, new LineController(line, true, this));
                        }
                    }
                }
                result.success(newIds);
                break;
            }
            case "line#addJson": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                List<Map<String, Object>> newSymbolIds = new ArrayList<Map<String, Object>>();
                final Object json = call.argument("json");
                if (json != null) {
                    List<Line> newSymbols = lineManager.create(Convert.toString(json));
                    String lineId;
                    for (Line line : newSymbols) {
                        lineId = String.valueOf(line.getId());
                        newSymbolIds.add(Convert.convertLineOptions(lineId, line));
                        lines.put(lineId, new LineController(line, annotationConsumeTapEvents.contains("AnnotationType.line"), this));
                    }
                }
                result.success(newSymbolIds);
                break;
            }
            case "line#removeAll": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                final ArrayList<String> ids = call.argument("ids");
                LineController lineController;

                List<Line> toBeRemoved = new ArrayList<Line>();
                for (String id : ids) {
                    lineController = lines.remove(id);
                    if (lineController != null) {
                        toBeRemoved.add(lineController.getLine());
                    }
                }
                if (!toBeRemoved.isEmpty()) {
                    lineManager.clear(toBeRemoved);
                }
                result.success(null);
                break;
            }
            case "line#update": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                final String lineId = call.argument("line");
                final LineController line = line(lineId);
                Convert.interpretLineOptions(call.argument("options"), line);
                line.update(lineManager);
                result.success(null);
                break;
            }
            case "line#getGeometry": {
                if (lineManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "line"), null);
                    return;
                }
                final String lineId = call.argument("line");
                final LineController line = line(lineId);
                final List<LatLng> lineLatLngs = line.getGeometry();
                final List<Object> resultList = new ArrayList<>();
                for (LatLng latLng : lineLatLngs) {
                    Map<String, Double> hashMapLatLng = new HashMap<>();
                    hashMapLatLng.put("latitude", latLng.getLatitude());
                    hashMapLatLng.put("longitude", latLng.getLongitude());
                    resultList.add(hashMapLatLng);
                }
                result.success(resultList);
                break;
            }
            case "circle#add": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                final CircleBuilder circleBuilder = newCircleBuilder();
                Convert.interpretCircleOptions(call.argument("options"), circleBuilder);
                final Circle circle = circleBuilder.build();
                final String circleId = String.valueOf(circle.getId());
                circles.put(circleId, new CircleController(circle, annotationConsumeTapEvents.contains("AnnotationType.circle"), this));
                result.success(circleId);
                break;
            }
            case "circle#addAll": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                List<String> newIds = new ArrayList<String>();
                final List<Object> options = call.argument("options");
                List<CircleOptions> optionList = new ArrayList<CircleOptions>();
                if (options != null) {
                    CircleBuilder builder;
                    for (Object o : options) {
                        builder = newCircleBuilder();
                        Convert.interpretCircleOptions(o, builder);
                        optionList.add(builder.getCircleOptions());
                    }
                    if (!optionList.isEmpty()) {
                        List<Circle> newCircles = circleManager.create(optionList);
                        String id;
                        for (Circle circle : newCircles) {
                            id = String.valueOf(circle.getId());
                            newIds.add(id);
                            circles.put(id, new CircleController(circle, true, this));
                        }
                    }
                }
                result.success(newIds);
                break;
            }
            case "circle#addJson": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                List<Map<String, Object>> newCircleIds = new ArrayList<Map<String, Object>>();
                final Object json = call.argument("json");
                if (json != null) {
                    List<Circle> newCircles = circleManager.create(Convert.toString(json));
                    String circleId;
                    for (Circle circle : newCircles) {
                        circleId = String.valueOf(circle.getId());
                        newCircleIds.add(Convert.convertCircleOptions(circleId, circle));
                        circles.put(circleId, new CircleController(circle, annotationConsumeTapEvents.contains("AnnotationType.circle"), this));
                    }
                }
                result.success(newCircleIds);
                break;
            }
            case "circle#removeAll": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                final ArrayList<String> ids = call.argument("ids");
                CircleController circleController;

                List<Circle> toBeRemoved = new ArrayList<Circle>();
                for (String id : ids) {
                    circleController = circles.remove(id);
                    if (circleController != null) {
                        toBeRemoved.add(circleController.getCircle());
                    }
                }
                if (!toBeRemoved.isEmpty()) {
                    circleManager.clear(toBeRemoved);
                }
                result.success(null);
                break;
            }
            case "circle#remove": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                final String circleId = call.argument("circle");
                removeCircle(circleId);
                result.success(null);
                break;
            }
            case "circle#update": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                Log.e(TAG, "update circle");
                final String circleId = call.argument("circle");
                final CircleController circle = circle(circleId);
                Convert.interpretCircleOptions(call.argument("options"), circle);
                circle.update(circleManager);
                result.success(null);
                break;
            }
            case "circle#getGeometry": {
                if (circleManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "circle"), null);
                    return;
                }
                final String circleId = call.argument("circle");
                final CircleController circle = circle(circleId);
                final LatLng circleLatLng = circle.getGeometry();
                Map<String, Double> hashMapLatLng = new HashMap<>();
                hashMapLatLng.put("latitude", circleLatLng.getLatitude());
                hashMapLatLng.put("longitude", circleLatLng.getLongitude());
                result.success(hashMapLatLng);
                break;
            }
            case "fill#add": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                final FillBuilder fillBuilder = newFillBuilder();
                Convert.interpretFillOptions(call.argument("options"), fillBuilder);
                final Fill fill = fillBuilder.build();
                final String fillId = String.valueOf(fill.getId());
                fills.put(fillId, new FillController(fill, annotationConsumeTapEvents.contains("AnnotationType.fill"), this));
                result.success(fillId);
                break;
            }

            case "fill#addAll": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                List<String> newIds = new ArrayList<String>();
                final List<Object> options = call.argument("options");
                List<FillOptions> optionList = new ArrayList<FillOptions>();
                if (options != null) {
                    FillBuilder builder;
                    for (Object o : options) {
                        builder = newFillBuilder();
                        Convert.interpretFillOptions(o, builder);
                        optionList.add(builder.getFillOptions());
                    }
                    if (!optionList.isEmpty()) {
                        List<Fill> newFills = fillManager.create(optionList);
                        String id;
                        for (Fill fill : newFills) {
                            id = String.valueOf(fill.getId());
                            newIds.add(id);
                            fills.put(id, new FillController(fill, true, this));
                        }
                    }
                }
                result.success(newIds);
                break;
            }
            case "fill#addJson": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                List<Map<String, Object>> newSymbolIds = new ArrayList<Map<String, Object>>();
                final Object json = call.argument("json");
                if (json != null) {
                    List<Fill> newFills = fillManager.create(Convert.toString(json));
                    String fillId;
                    for (Fill line : newFills) {
                        fillId = String.valueOf(line.getId());
                        newSymbolIds.add(Convert.convertFillOptions(fillId, line));
                        fills.put(fillId, new FillController(line, annotationConsumeTapEvents.contains("AnnotationType.fill"), this));
                    }
                }
                result.success(newSymbolIds);
                break;
            }
            case "fill#removeAll": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                final ArrayList<String> ids = call.argument("ids");
                FillController fillController;

                List<Fill> toBeRemoved = new ArrayList<Fill>();
                for (String id : ids) {
                    fillController = fills.remove(id);
                    if (fillController != null) {
                        toBeRemoved.add(fillController.getFill());
                    }
                }
                if (!toBeRemoved.isEmpty()) {
                    fillManager.clear(toBeRemoved);
                }
                result.success(null);
                break;
            }
            case "fill#remove": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                final String fillId = call.argument("fill");
                removeFill(fillId);
                result.success(null);
                break;
            }
            case "fill#update": {
                if (fillManager == null) {
                    result.error(annotationManagerNotCreatedErrorCode, String.format(annotationManagerNotCreatedErrorCode, "fill"), null);
                    return;
                }
                final String fillId = call.argument("fill");
                final FillController fill = fill(fillId);
                Convert.interpretFillOptions(call.argument("options"), fill);
                fill.update(fillManager);
                result.success(null);
                break;
            }
            case "source#addGeoJson": {
                final String sourceId = call.argument("sourceId");
                final String geojson = call.argument("geojson");
                addGeoJsonSource(sourceId, geojson);
                result.success(null);
                break;
            }
            case "source#setGeoJson": {
                final String sourceId = call.argument("sourceId");
                final String geojson = call.argument("geojson");
                setGeoJsonSource(sourceId, geojson);
                result.success(null);
                break;
            }
            case "symbolLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final String sourceLayer = call.argument("sourceLayer");
                final PropertyValue[] properties = LayerPropertyConverter.interpretSymbolLayerProperties(call.argument("properties"));
                addSymbolLayer(layerId, sourceId, belowLayerId, sourceLayer, properties, null);
                result.success(null);
                break;
            }
            case "lineLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final String sourceLayer = call.argument("sourceLayer");
                final PropertyValue[] properties = LayerPropertyConverter.interpretLineLayerProperties(call.argument("properties"));
                addLineLayer(layerId, sourceId, belowLayerId, sourceLayer, properties, null);
                result.success(null);
                break;
            }
            case "fillLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final String sourceLayer = call.argument("sourceLayer");
                final PropertyValue[] properties = LayerPropertyConverter.interpretFillLayerProperties(call.argument("properties"));
                addFillLayer(layerId, sourceId, belowLayerId, sourceLayer, properties, null);
                result.success(null);
                break;
            }
            case "circleLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final String sourceLayer = call.argument("sourceLayer");
                final PropertyValue[] properties = LayerPropertyConverter.interpretCircleLayerProperties(call.argument("properties"));
                addCircleLayer(layerId, sourceId, belowLayerId, sourceLayer, properties, null);
                result.success(null);
                break;
            }
            case "rasterLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final PropertyValue[] properties = LayerPropertyConverter.interpretRasterLayerProperties(call.argument("properties"));
                addRasterLayer(layerId, sourceId, belowLayerId, properties, null);
                result.success(null);
                break;
            }
            case "hillshadeLayer#add": {
                final String sourceId = call.argument("sourceId");
                final String layerId = call.argument("layerId");
                final String belowLayerId = call.argument("belowLayerId");
                final PropertyValue[] properties = LayerPropertyConverter.interpretHillshadeLayerProperties(call.argument("properties"));
                addHillshadeLayer(layerId, sourceId, belowLayerId, properties, null);
                result.success(null);
                break;
            }
            case "locationComponent#getLastLocation": {
                Log.e(TAG, "location component: getLastLocation");
                if (this.myLocationEnabled && locationComponent != null && locationEngine != null) {
                    Map<String, Object> reply = new HashMap<>();
                    locationEngine.getLastLocation(new LocationEngineCallback<LocationEngineResult>() {
                        @Override
                        public void onSuccess(LocationEngineResult locationEngineResult) {
                            Location lastLocation = locationEngineResult.getLastLocation();
                            if (lastLocation != null) {
                                reply.put("latitude", lastLocation.getLatitude());
                                reply.put("longitude", lastLocation.getLongitude());
                                reply.put("altitude", lastLocation.getAltitude());
                                result.success(reply);
                            } else {
                                result.error("", "", null); // ???
                            }
                        }

                        @Override
                        public void onFailure(@NonNull Exception exception) {
                            result.error("", "", null); // ???
                        }
                    });
                }
                break;
            }
            case "style#addImage": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                style.addImage(call.argument("name"), BitmapFactory.decodeByteArray(call.argument("bytes"), 0, call.argument("length")), call.argument("sdf"));
                result.success(null);
                break;
            }
            case "style#addImageSource": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                List<LatLng> coordinates = Convert.toLatLngList(call.argument("coordinates"), false);
                style.addSource(new ImageSource(call.argument("imageSourceId"), new LatLngQuad(coordinates.get(0), coordinates.get(1), coordinates.get(2), coordinates.get(3)), BitmapFactory.decodeByteArray(call.argument("bytes"), 0, call.argument("length"))));
                result.success(null);
                break;
            }
            case "style#addSource": {
                final String id = Convert.toString(call.argument("sourceId"));
                final Map<String, Object> properties = (Map<String, Object>) call.argument("properties");
                SourcePropertyConverter.addSource(id, properties, style);
                result.success(null);
                break;
            }

            case "style#removeSource": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                style.removeSource((String) call.argument("sourceId"));
                result.success(null);
                break;
            }
            case "style#setSource": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                style.removeSource((String) call.argument("sourceId"));
                result.success(null);
                break;
            }
            case "style#addLayer": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                style.addLayer(new RasterLayer(call.argument("imageLayerId"), call.argument("imageSourceId")));
                result.success(null);
                break;
            }
            case "style#addLayerBelow": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                style.addLayerBelow(new RasterLayer(call.argument("imageLayerId"), call.argument("imageSourceId")), call.argument("belowLayerId"));
                result.success(null);
                break;
            }
            case "style#removeLayer": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                String layerId = call.argument("layerId");
                style.removeLayer(layerId);
                featureLayerIdentifiers.remove(layerId);

                result.success(null);
                break;
            }
            case "style#showGeoAnalytics": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                MapmyIndiaGeoAnalyticsType analyticsType = Convert.getGeoAnalyticsType(Convert.toInt(call.argument("geoAnalyticsType")));
                List<MapmyIndiaGeoAnalyticsRequest> requests = Convert.getGeoAnalyticsRequest(call.arguments);
                geoAnalyticsPlugin.showGeoAnalytics(analyticsType, requests);
                result.success(null);
                break;
            }
            case "style#removeGeoAnalytics": {
                if (style == null) {
                    result.error("STYLE IS NULL", "The style is null. Has onStyleLoaded() already been invoked?", null);
                }
                MapmyIndiaGeoAnalyticsType analyticsType = Convert.getGeoAnalyticsType(Convert.toInt(call.argument("geoAnalyticsType")));
                geoAnalyticsPlugin.removeGeoAnalytics(analyticsType);
                result.success(null);
                break;
            }
            case "map#enableTraffic": {
                final Boolean enable = Convert.toBoolean(call.arguments);
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                mapmyIndiaMap.enableTraffic(enable);
                result.success(null);
                break;
            }
            case "map#enableTrafficClosure": {
                final Boolean enable = Convert.toBoolean(call.arguments);
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                mapmyIndiaMap.enableTrafficClosure(enable);
                result.success(null);
                break;
            }
            case "map#enableTrafficFreeFlow": {
                final Boolean enable = Convert.toBoolean(call.arguments);
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                mapmyIndiaMap.enableTrafficFreeFlow(enable);
                result.success(null);
                break;
            }
            case "map#enableTrafficNonFreeFlow": {
                final Boolean enable = Convert.toBoolean(call.arguments);
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                mapmyIndiaMap.enableTrafficNonFreeFlow(enable);
                result.success(null);
                break;
            }
            case "map#enableTrafficStopIcon": {
                final Boolean enable = Convert.toBoolean(call.arguments);
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                mapmyIndiaMap.enableTrafficStopIcon(enable);
                result.success(null);
                break;
            }
            case "map#isEnableTraffic": {
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                result.success(mapmyIndiaMap.isEnableTraffic());
                break;
            }
            case "map#isEnableTrafficClosure": {
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                result.success(mapmyIndiaMap.isEnableTrafficClosure());
                break;
            }
            case "map#isEnableTrafficFreeFlow": {
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                result.success(mapmyIndiaMap.isEnableTrafficFreeFlow());
                break;
            }
            case "map#isEnableTrafficNonFreeFlow": {
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                result.success(mapmyIndiaMap.isEnableTrafficNonFreeFlow());
                break;
            }
            case "map#isEnableTrafficStopIcon": {
                if (mapmyIndiaMap == null) {
                    result.error("MAP IS NULL", "The map is null. Has onMapReady() already been invoked?", null);
                }
                result.success(mapmyIndiaMap.isEnableTrafficStopIcon());
                break;
            }
            default:
                result.notImplemented();
        }
    }

    @Override
    public void onCameraMoveStarted(int reason) {
        final Map<String, Object> arguments = new HashMap<>(2);
        boolean isGesture = reason == MapmyIndiaMap.OnCameraMoveStartedListener.REASON_API_GESTURE;
        arguments.put("isGesture", isGesture);
        methodChannel.invokeMethod("camera#onMoveStarted", arguments);
    }

    @Override
    public void onCameraMove() {
        if (!trackCameraPosition) {
            return;
        }
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("position", Convert.toJson(mapmyIndiaMap.getCameraPosition()));
        methodChannel.invokeMethod("camera#onMove", arguments);
    }

    @Override
    public void onCameraIdle() {
        final Map<String, Object> arguments = new HashMap<>(2);
        if (trackCameraPosition) {
            arguments.put("position", Convert.toJson(mapmyIndiaMap.getCameraPosition()));
        }
        methodChannel.invokeMethod("camera#onIdle", arguments);
    }

    @Override
    public void onCameraTrackingChanged(int currentMode) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("mode", currentMode);
        methodChannel.invokeMethod("map#onCameraTrackingChanged", arguments);
    }

    @Override
    public void onCameraTrackingDismissed() {
        this.myLocationTrackingMode = 0;
        methodChannel.invokeMethod("map#onCameraTrackingDismissed", new HashMap<>());
    }

    @Override
    public void onDidBecomeIdle() {
        methodChannel.invokeMethod("map#onIdle", new HashMap<>());
    }

    @Override
    public boolean onAnnotationClick(Annotation annotation) {
        if (annotation instanceof Symbol) {
            final SymbolController symbolController = symbols.get(String.valueOf(annotation.getId()));
            if (symbolController != null) {
                return symbolController.onTap();
            }
        }

        if (annotation instanceof Line) {
            final LineController lineController = lines.get(String.valueOf(annotation.getId()));
            if (lineController != null) {
                return lineController.onTap();
            }
        }

        if (annotation instanceof Circle) {
            final CircleController circleController = circles.get(String.valueOf(annotation.getId()));
            if (circleController != null) {
                return circleController.onTap();
            }
        }
        if (annotation instanceof Fill) {
            final FillController fillController = fills.get(String.valueOf(annotation.getId()));
            if (fillController != null) {
                return fillController.onTap();
            }
        }
        return false;
    }

    @Override
    public void onSymbolTapped(Symbol symbol) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("symbol", String.valueOf(symbol.getId()));
        methodChannel.invokeMethod("symbol#onTap", arguments);
    }

    @Override
    public void onLineTapped(Line line) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("line", String.valueOf(line.getId()));
        methodChannel.invokeMethod("line#onTap", arguments);
    }

    @Override
    public void onCircleTapped(Circle circle) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("circle", String.valueOf(circle.getId()));
        methodChannel.invokeMethod("circle#onTap", arguments);
    }

    @Override
    public void onFillTapped(Fill fill) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("fill", String.valueOf(fill.getId()));
        methodChannel.invokeMethod("fill#onTap", arguments);
    }

    @Override
    public boolean onMapClick(@NonNull LatLng point) {
        PointF pointf = mapmyIndiaMap.getProjection().toScreenLocation(point);
        RectF rectF = new RectF(
                pointf.x - 10,
                pointf.y - 10,
                pointf.x + 10,
                pointf.y + 10
        );
        Feature feature = firstFeatureOnLayers(rectF);
        final Map<String, Object> arguments = new HashMap<>();
        arguments.put("x", pointf.x);
        arguments.put("y", pointf.y);
        arguments.put("lng", point.getLongitude());
        arguments.put("lat", point.getLatitude());
        if (feature != null) {
            arguments.put("id", feature.id());
            methodChannel.invokeMethod("feature#onTap", arguments);
        } else {
            methodChannel.invokeMethod("map#onMapClick", arguments);
        }
        return false;
    }

    @Override
    public boolean onMapLongClick(@NonNull LatLng point) {
        PointF pointf = mapmyIndiaMap.getProjection().toScreenLocation(point);
        final Map<String, Object> arguments = new HashMap<>(5);
        arguments.put("x", pointf.x);
        arguments.put("y", pointf.y);
        arguments.put("lng", point.getLongitude());
        arguments.put("lat", point.getLatitude());
        methodChannel.invokeMethod("map#onMapLongClick", arguments);
        return true;
    }

    @Override
    public void dispose() {
        if (disposed) {
            return;
        }
        disposed = true;
        methodChannel.setMethodCallHandler(null);

        mapView.onResume();
        mapView.onStop();
        destroyMapViewIfNecessary();
        Lifecycle lifecycle = lifecycleProvider.getLifecycle();
        if (lifecycle != null) {
            lifecycle.removeObserver(this);
        }
    }

    private void destroyMapViewIfNecessary() {
        if (mapView == null) {
            return;
        }

        if (locationComponent != null) {
            locationComponent.setLocationComponentEnabled(false);
        }
        if (symbolManager != null) {
            symbolManager.onDestroy();
        }
        if (lineManager != null) {
            lineManager.onDestroy();
        }
        if (circleManager != null) {
            circleManager.onDestroy();
        }
        if (fillManager != null) {
            fillManager.onDestroy();
        }
        stopListeningForLocationUpdates();

        mapView.onDestroy();
        mapView = null;
    }

    @Override
    public void onCreate(@NonNull LifecycleOwner owner) {
        if (disposed) {
            return;
        }
        mapView.onCreate(null);
    }

    @Override
    public void onStart(@NonNull LifecycleOwner owner) {
        if (disposed) {
            return;
        }
        mapView.onStart();
    }

    @Override
    public void onResume(@NonNull LifecycleOwner owner) {
        if (disposed) {
            return;
        }
        mapView.onResume();
        if (myLocationEnabled) {
            startListeningForLocationUpdates();
        }
    }

    @Override
    public void onPause(@NonNull LifecycleOwner owner) {
        if (disposed) {
            return;
        }
        mapView.onPause();
    }

    @Override
    public void onStop(@NonNull LifecycleOwner owner) {
        if (disposed) {
            return;
        }
        mapView.onStop();
    }

    @Override
    public void onDestroy(@NonNull LifecycleOwner owner) {
        owner.getLifecycle().removeObserver(this);

        methodChannel.setMethodCallHandler(null);
        if (disposed) {
            return;
        }
        destroyMapViewIfNecessary();
    }

    // MapboxMapOptionsSink methods

    @Override
    public void setCameraTargetBounds(LatLngBounds bounds) {
        this.bounds = bounds;
    }

    @Override
    public void setCompassEnabled(boolean compassEnabled) {
        mapmyIndiaMap.getUiSettings().setCompassEnabled(compassEnabled);
    }

    @Override
    public void setTrackCameraPosition(boolean trackCameraPosition) {
        this.trackCameraPosition = trackCameraPosition;
    }

    @Override
    public void setRotateGesturesEnabled(boolean rotateGesturesEnabled) {
        mapmyIndiaMap.getUiSettings().setRotateGesturesEnabled(rotateGesturesEnabled);
    }

    @Override
    public void setScrollGesturesEnabled(boolean scrollGesturesEnabled) {
        mapmyIndiaMap.getUiSettings().setScrollGesturesEnabled(scrollGesturesEnabled);
    }

    @Override
    public void setTiltGesturesEnabled(boolean tiltGesturesEnabled) {
        mapmyIndiaMap.getUiSettings().setTiltGesturesEnabled(tiltGesturesEnabled);
    }

    @Override
    public void setMinMaxZoomPreference(Float min, Float max) {
        mapmyIndiaMap.setMinZoomPreference(min != null ? min : MapmyIndiaConstants.MINIMUM_ZOOM);
        mapmyIndiaMap.setMaxZoomPreference(max != null ? max : MapmyIndiaConstants.MAXIMUM_ZOOM);
    }

    @Override
    public void setZoomGesturesEnabled(boolean zoomGesturesEnabled) {
        mapmyIndiaMap.getUiSettings().setZoomGesturesEnabled(zoomGesturesEnabled);
    }

    @Override
    public void setMyLocationEnabled(boolean myLocationEnabled) {
        if (this.myLocationEnabled == myLocationEnabled) {
            return;
        }
        this.myLocationEnabled = myLocationEnabled;
        if (mapmyIndiaMap != null) {
            updateMyLocationEnabled();
        }
    }

    @Override
    public void setMyLocationTrackingMode(int myLocationTrackingMode) {
        if (this.myLocationTrackingMode == myLocationTrackingMode) {
            return;
        }
        this.myLocationTrackingMode = myLocationTrackingMode;
        if (mapmyIndiaMap != null && locationComponent != null) {
            updateMyLocationTrackingMode();
        }
    }

    @Override
    public void setMyLocationRenderMode(int myLocationRenderMode) {
        if (this.myLocationRenderMode == myLocationRenderMode) {
            return;
        }
        this.myLocationRenderMode = myLocationRenderMode;
        if (mapmyIndiaMap != null && locationComponent != null) {
            updateMyLocationRenderMode();
        }
    }

    public void setLogoViewMargins(int x, int y) {
        mapmyIndiaMap.getUiSettings().setLogoMargins(x, 0, 0, y);
    }

    @Override
    public void setCompassGravity(int gravity) {
        switch (gravity) {
            case 0:
                mapmyIndiaMap.getUiSettings().setCompassGravity(Gravity.TOP | Gravity.START);
                break;
            default:
            case 1:
                mapmyIndiaMap.getUiSettings().setCompassGravity(Gravity.TOP | Gravity.END);
                break;
            case 2:
                mapmyIndiaMap.getUiSettings().setCompassGravity(Gravity.BOTTOM | Gravity.START);
                break;
            case 3:
                mapmyIndiaMap.getUiSettings().setCompassGravity(Gravity.BOTTOM | Gravity.END);
                break;
        }
    }

    @Override
    public void setCompassViewMargins(int x, int y) {
        switch (mapmyIndiaMap.getUiSettings().getCompassGravity()) {
            case Gravity.TOP | Gravity.START:
                mapmyIndiaMap.getUiSettings().setCompassMargins(x, y, 0, 0);
                break;
            default:
            case Gravity.TOP | Gravity.END:
                mapmyIndiaMap.getUiSettings().setCompassMargins(0, y, x, 0);
                break;
            case Gravity.BOTTOM | Gravity.START:
                mapmyIndiaMap.getUiSettings().setCompassMargins(x, 0, 0, y);
                break;
            case Gravity.BOTTOM | Gravity.END:
                mapmyIndiaMap.getUiSettings().setCompassMargins(0, 0, x, y);
                break;
        }
    }

    @Override
    public void setAttributionButtonGravity(int gravity) {
        switch (gravity) {
            case 0:
                mapmyIndiaMap.getUiSettings().setAttributionGravity(Gravity.TOP | Gravity.START);
                break;
            default:
            case 1:
                mapmyIndiaMap.getUiSettings().setAttributionGravity(Gravity.TOP | Gravity.END);
                break;
            case 2:
                mapmyIndiaMap.getUiSettings().setAttributionGravity(Gravity.BOTTOM | Gravity.START);
                break;
            case 3:
                mapmyIndiaMap.getUiSettings().setAttributionGravity(Gravity.BOTTOM | Gravity.END);
                break;
        }
    }

    @Override
    public void setAttributionButtonMargins(int x, int y) {
        switch (mapmyIndiaMap.getUiSettings().getAttributionGravity()) {
            case Gravity.TOP | Gravity.START:
                mapmyIndiaMap.getUiSettings().setAttributionMargins(x, y, 0, 0);
                break;
            default:
            case Gravity.TOP | Gravity.END:
                mapmyIndiaMap.getUiSettings().setAttributionMargins(0, y, x, 0);
                break;
            case Gravity.BOTTOM | Gravity.START:
                mapmyIndiaMap.getUiSettings().setAttributionMargins(x, 0, 0, y);
                break;
            case Gravity.BOTTOM | Gravity.END:
                mapmyIndiaMap.getUiSettings().setAttributionMargins(0, 0, x, y);
                break;
        }
    }

    @Override
    public void showGeoAnalyticsInfoWindow(boolean show) {
        this.showGeoAnalyticsInfoWindow = show;
        if(geoAnalyticsPlugin != null) {
            geoAnalyticsPlugin.shouldShowInfoWindow(show);
        }
    }

    private void updateMyLocationEnabled() {
        if (this.locationComponent == null && myLocationEnabled) {
            enableLocationComponent(mapmyIndiaMap.getStyle());
        }

        if (myLocationEnabled) {
            startListeningForLocationUpdates();
        } else {
            stopListeningForLocationUpdates();
        }

        locationComponent.setLocationComponentEnabled(myLocationEnabled);
    }

    private void startListeningForLocationUpdates() {
        if (locationEngineCallback == null && locationComponent != null && locationComponent.getLocationEngine() != null) {
            locationEngineCallback = new LocationEngineCallback<LocationEngineResult>() {
                @Override
                public void onSuccess(LocationEngineResult result) {
                    onUserLocationUpdate(result.getLastLocation());
                }

                @Override
                public void onFailure(@NonNull Exception exception) {
                }
            };
            locationComponent.getLocationEngine().requestLocationUpdates(locationComponent.getLocationEngineRequest(), locationEngineCallback, null);
        }
    }

    private void stopListeningForLocationUpdates() {
        if (locationEngineCallback != null && locationComponent != null && locationComponent.getLocationEngine() != null) {
            locationComponent.getLocationEngine().removeLocationUpdates(locationEngineCallback);
            locationEngineCallback = null;
        }
    }

    private void updateMyLocationTrackingMode() {
        int[] mapboxTrackingModes = new int[]{CameraMode.NONE, CameraMode.TRACKING, CameraMode.TRACKING_COMPASS, CameraMode.TRACKING_GPS, CameraMode.TRACKING_GPS_NORTH, CameraMode.NONE_COMPASS, CameraMode.NONE_GPS};
        locationComponent.setCameraMode(mapboxTrackingModes[this.myLocationTrackingMode]);
    }

    private void updateMyLocationRenderMode() {
        int[] mapboxRenderModes = new int[]{RenderMode.NORMAL, RenderMode.COMPASS, RenderMode.GPS};
        locationComponent.setRenderMode(mapboxRenderModes[this.myLocationRenderMode]);
    }

    private boolean hasLocationPermission() {
        return checkSelfPermission(Manifest.permission.ACCESS_FINE_LOCATION)
                == PackageManager.PERMISSION_GRANTED
                || checkSelfPermission(Manifest.permission.ACCESS_COARSE_LOCATION)
                == PackageManager.PERMISSION_GRANTED;
    }

    private int checkSelfPermission(String permission) {
        if (permission == null) {
            throw new IllegalArgumentException("permission is null");
        }
        return context.checkPermission(
                permission, android.os.Process.myPid(), android.os.Process.myUid());
    }

    /**
     * Tries to find highest scale image for display type
     *
     * @param imageId
     * @param density
     * @return
     */
    private Bitmap getScaledImage(String imageId, float density) {
        AssetFileDescriptor assetFileDescriptor;

        // Split image path into parts.
        List<String> imagePathList = Arrays.asList(imageId.split("/"));
        List<String> assetPathList = new ArrayList<>();

        // "On devices with a device pixel ratio of 1.8, the asset .../2.0x/my_icon.png would be chosen.
        // For a device pixel ratio of 2.7, the asset .../3.0x/my_icon.png would be chosen."
        // Source: https://flutter.dev/docs/development/ui/assets-and-images#resolution-aware
        for (int i = (int) Math.ceil(density); i > 0; i--) {
            String assetPath;
            if (i == 1) {
                // If density is 1.0x then simply take the default asset path
                assetPath = MapboxMapsPlugin.flutterAssets.getAssetFilePathByName(imageId);
            } else {
                // Build a resolution aware asset path as follows:
                // <directory asset>/<ratio>/<image name>
                // where ratio is 1.0x, 2.0x or 3.0x.
                StringBuilder stringBuilder = new StringBuilder();
                for (int j = 0; j < imagePathList.size() - 1; j++) {
                    stringBuilder.append(imagePathList.get(j));
                    stringBuilder.append("/");
                }
                stringBuilder.append(((float) i) + "x");
                stringBuilder.append("/");
                stringBuilder.append(imagePathList.get(imagePathList.size() - 1));
                assetPath = MapboxMapsPlugin.flutterAssets.getAssetFilePathByName(stringBuilder.toString());
            }
            // Build up a list of resolution aware asset paths.
            assetPathList.add(assetPath);
        }

        // Iterate over asset paths and get the highest scaled asset (as a bitmap).
        Bitmap bitmap = null;
        for (String assetPath : assetPathList) {
            try {
                // Read path (throws exception if doesn't exist).
                assetFileDescriptor = mapView.getContext().getAssets().openFd(assetPath);
                InputStream assetStream = assetFileDescriptor.createInputStream();
                bitmap = BitmapFactory.decodeStream(assetStream);
                assetFileDescriptor.close(); // Close for memory
                break; // If exists, break
            } catch (IOException e) {
                // Skip
            }
        }
        return bitmap;
    }

    @Override
    public void getSelectedFeatures(List<MapmyIndiaGeoAnalyticsDetail> list) {
        mapView.post(new Runnable() {
            @Override
            public void run() {
                final Map<String, Object> arguments = new HashMap<>();
                arguments.put("selectedFeatures", Convert.getGeoAnalyticsDetails(list));
                methodChannel.invokeMethod("geoanalytics#onSelectedFeatures", arguments);
            }
        });

    }

    /**
     * Simple Listener to listen for the status of camera movements.
     */
    public class OnCameraMoveFinishedListener implements MapmyIndiaMap.CancelableCallback {
        @Override
        public void onFinish() {
        }

        @Override
        public void onCancel() {
        }
    }
}
