import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:mapmyindia_gl_example/utils/color.dart';
import 'package:mapmyindia_gl/mapmyindia_gl.dart';

class HateOsNearbyWidget extends StatefulWidget {
  @override
  State<StatefulWidget> createState() {
    return HateOsNearbyWidgetState();
  }
}

class HateOsNearbyWidgetState extends State {
  static final CameraPosition _kInitialPosition = const CameraPosition(
    target: LatLng(25.321684, 82.987289),
    zoom: 14.0,
  );

  List<NearbyResult> result = [];

  bool isShowList = false;

  late MapmyIndiaMapController controller;
  List<SuggestedSearch> _suggestedSearches = [];
  bool showAutoSuggestWidget = false;
  TextEditingController textController = TextEditingController();

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        centerTitle: true,
        backgroundColor: MyColor.colorPrimary,
        brightness: Brightness.dark,
        title: Text(
          'Nearby API',
          style: TextStyle(color: Colors.white),
        ),
        elevation: 0.2,
      ),
      body: Stack(children: <Widget>[
        MapmyIndiaMap(
          initialCameraPosition: _kInitialPosition,
          onMapCreated: (map) =>
          {
            controller = map,
          },
          onStyleLoadedCallback: () => {
            setState(() {
              showAutoSuggestWidget = true;
            })
          },
        ),
        showAutoSuggestWidget ? autosuggestWidget(): Container(),
        result.length > 0 && isShowList
            ? BottomSheet(
            onClosing: () => {},

            builder: (context) =>
                Expanded(
                    child: ListView.builder(
                        itemCount: result.length,
                        itemBuilder: (context, index) {
                          return Container(
                              padding: EdgeInsets.fromLTRB(12, 0, 12, 0),
                              decoration: BoxDecoration(
                                color: Colors.white,
                              ),
                              child: ListTile(
                                contentPadding: EdgeInsets.all(5),
                                focusColor: Colors.white,
                                title: Text(result[index].placeName ?? ''),
                                subtitle: Text(
                                  result[index].placeAddress ?? '',
                                  maxLines: 2,
                                ),
                              ));
                        })))
            : Container()
      ]),
      floatingActionButton: FloatingActionButton(
        onPressed: () =>
        {
          setState(() {
            isShowList = !isShowList;
          })
        },
        child: isShowList ? Icon(Icons.map) : Icon(Icons.list),
      ),
    );
  }


  Widget autosuggestWidget() {
    return Column(
      children: [
        SizedBox(
          height: 20,
        ),
        Container(
            padding: EdgeInsets.fromLTRB(12, 0, 12, 0),
            decoration: BoxDecoration(
              color: Colors.white,
            ),
            child: TextFormField(
                controller: textController,
                decoration: InputDecoration(
                    hintText: "Search Here", fillColor: Colors.white),
                maxLines: 1,
                onChanged: (text) {
                  callAutoSuggest(text);
                })),
        SizedBox(
          height: 20,
        ),
        _suggestedSearches.length > 0
            ? Expanded(
            child: Container(
                padding: EdgeInsets.fromLTRB(12, 0, 12, 0),
                decoration: BoxDecoration(
                  color: Colors.white,
                ),
                child: ListView.builder(
                    itemCount: _suggestedSearches.length,
                    itemBuilder: (context, index) {
                      return ListTile(
                        contentPadding: EdgeInsets.all(12),
                        focusColor: Colors.white,
                        title: Text('${_suggestedSearches[index].keyword ?? ''} ${_suggestedSearches[index].identifier ?? ''} ${_suggestedSearches[index].location ?? ''}'),
                        onTap: () {
                          nearby(_suggestedSearches[index]);
                          setState(() {
                            _suggestedSearches = [];
                          });
                          textController.text = '';
                        },
                      );
                    })))
            : Container()
      ],
    );
  }

  Future<void> addImageFromAsset(String name, String assetName) async {
    final ByteData bytes = await rootBundle.load(assetName);
    final Uint8List list = bytes.buffer.asUint8List();
    return controller.addImage(name, list);
  }

  void callAutoSuggest(String text) async {
    if (text.length > 2) {
      try {
        AutoSuggestResponse? response =
        await MapmyIndiaAutoSuggest(query: text, bridge: true).callAutoSuggest();
        if (response?.suggestedSearches != null) {
          setState(() {
            _suggestedSearches = response!.suggestedSearches!;
          });
        } else {
          setState(() {
            _suggestedSearches = [];
          });
        }
      } catch (e) {
        if (e is PlatformException) {
          Fluttertoast.showToast(msg: '${e.code} --- ${e.message}');
        }
      }
    } else {
      setState(() {
        _suggestedSearches = [];
      });
    }
  }

  boundsFromLatLngList(List<LatLng> list) {
    assert(list.isNotEmpty);
    double? x0, x1, y0, y1;
    for (LatLng latLng in list) {
      if (x0 == null || x1 == null || y0 == null || y1 == null) {
        x0 = x1 = latLng.latitude;
        y0 = y1 = latLng.longitude;
      } else {
        if (latLng.latitude > x1) x1 = latLng.latitude;
        if (latLng.latitude < x0) x0 = latLng.latitude;
        if (latLng.longitude > y1) y1 = latLng.longitude;
        if (latLng.longitude < y0) y0 = latLng.longitude;
      }
    }
    return LatLngBounds(
        northeast: LatLng(x1!, y1!), southwest: LatLng(x0!, y0!));
  }

  nearby(SuggestedSearch searches) async {
    await addImageFromAsset("icon", "assets/symbols/custom-icon.png");
    await controller.clearSymbols();
    setState(() {
      result =[];
    });
    if(searches.hyperLink == null) {
      return;
    }
    try {
        NearbyResponse? nearbyResponse = await MapmyIndiaHateOsNearby(searches.hyperLink!).callNearby();
        getResponse(nearbyResponse);
    } catch (e) {
      if (e is PlatformException) {
        Fluttertoast.showToast(msg: '${e.code} --- ${e.message}');
      } else {
        print(e);
      }
    }
  }

  void getResponse(NearbyResponse? nearbyResponse) {
    if (nearbyResponse != null &&
        nearbyResponse.suggestedLocations != null &&
        nearbyResponse.suggestedLocations!.length > 0) {
      // print(nearbyResponse.toJson());
      List<SymbolOptions> options = [];
      List<LatLng> latlngList = [];
      nearbyResponse.suggestedLocations!.forEach((element) {
        print(json.encode(element.toJson()));
        latlngList.add(LatLng(element.latitude!, element.longitude!));
        options.add(SymbolOptions(
          geometry: LatLng(element.latitude!, element.longitude!),
        ));
      });
      setState(() {
        result = nearbyResponse.suggestedLocations!;
      });
      controller.addSymbols(options);
      controller.animateCamera(
          CameraUpdate.newLatLngBounds(boundsFromLatLngList(latlngList), top: 40));
    }
  }
}
