import 'package:mapmyindia_gl_example/model/heatmap_option.dart';
import 'package:mapmyindia_gl_example/utils/color.dart';
import 'package:mapmyindia_gl/mapmyindia_gl.dart';
import 'package:flutter/material.dart';

class HeatMapWidget extends StatefulWidget {
  @override
  State<StatefulWidget> createState() {
    return HeatMapWidgetState();
  }
}

class HeatMapWidgetState extends State {
  late MapmyIndiaMapController mapController;
  static final CameraPosition _kInitialPosition = const CameraPosition(
    target: LatLng(25.321684, 82.987289),
    zoom: 14.0,
  );

  final List<HeatMapOption> heatMapOptions = [
    HeatMapOption(LatLng(28.101, 77.5129), 2.3),
    HeatMapOption(LatLng(28.1021, 77.5132), 2.0),
    HeatMapOption(LatLng(28.1224, 76.4048), 1.7),
    HeatMapOption(LatLng(28.0781, 77.3597), 1.6),
    HeatMapOption(LatLng(28.0781, 77.3597), 1.6),
    HeatMapOption(LatLng(28.1725, 74.789), 2.4),
    HeatMapOption(LatLng(28.0879, 77.512), 1.8),
    HeatMapOption(LatLng(28.674242, 74.2832), 1.3),
    HeatMapOption(LatLng(24.6778728, 80.8244288), 2.4),
    HeatMapOption(LatLng(24.6778728, 81.637417), 3.2),
    HeatMapOption(LatLng(27.8302397, 77.7372706), 1.2),
  ];

  // HeatMapOption(Point.fromLngLat(79.2973292, 28.3729432), 4.2));
  // HeatMapOption(Point.fromLngLat(79.7477686, 26.2351935), 2.13));
  // HeatMapOption(Point.fromLngLat(80.1762354, 26.1760509), 3.2));
  // HeatMapOption(Point.fromLngLat(79.7367823, 27.333618), 1.2));
  // HeatMapOption(Point.fromLngLat(80.9452784, 26.628706), 5.2));
  // HeatMapOption(Point.fromLngLat(78.6930811, 28.9032672), 3.2));
  // HeatMapOption(Point.fromLngLat(78.73702641, 29.3255866), 4.0));
  // HeatMapOption(Point.fromLngLat(79.4181788, 29.5838759), 5.0));

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        centerTitle: true,
        backgroundColor: MyColor.colorPrimary,
        brightness: Brightness.dark,
        title: Text(
          'Camera Feature',
          style: TextStyle(color: Colors.white),
        ),
        elevation: 0.2,
      ),
      body: MapmyIndiaMap(
        initialCameraPosition: _kInitialPosition,
        onMapCreated: (mapController) => this.mapController = mapController,
        onStyleLoadedCallback: _onStyleLoadedCallback,
      ),
    );
  }

  _onStyleLoadedCallback() async {
    await mapController.addSource("heat-map-source-id", GeojsonSourceProperties(data: getHeatMapData()));

    // await mapController.addHea
  }

  Map<String, dynamic> getHeatMapData() {
    Map<String, dynamic> featureCollection = {"type": "FeatureCollection"};
    List<dynamic> features = [];
    heatMapOptions.forEach((element) {
      features.add({
        "type": "Feature",
        "properties": {"mag": element.mag},
        "geometry": {
          "type": "Point",
          "coordinates": [element.latLng.longitude, element.latLng.latitude]
        }
      });
    });
    featureCollection["features"] = features;
    return featureCollection;
  }
}
