//
//  RestAPIMethodCall.swift
//  mapmyindia_gl
//
//  Created by CEINFO on 20/05/21.
//

import Foundation
import MapmyIndiaAPIKit
import MapmyIndiaDirections
import Flutter
import MapmyIndiaGeoanalytics

class RestAPIMethodCall: NSObject {
    
    init(call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "restApi#autosuggest":
            let autoSuggestManager = MapmyIndiaAutoSuggestManager.shared
            let option = RestAPIHelper.autosuggestOption(argument: call.arguments)
            if(option != nil) {
                autoSuggestManager.getAutoSuggestionResults(option!){ (suggestions, error) in
                if let error = error {
                    result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                } else {
                    let response = suggestions as? MapmyindiaAutoSuggestLocationResults
                    let theJSONText = RestAPIHelper.autosuggestResponse(response: response)
                    result("\(theJSONText ?? "")")
                }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#geocode":
            let geocodeManager = MapmyIndiaAtlasGeocodeManager.shared
            if let option = RestAPIHelper.geocodeOption(argument: call.arguments) {
                geocodeManager.getGeocodeResults(option) {(response, error) in
                if let error = error {
                    result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                } else {
                    let theJSONText = RestAPIHelper.geocodeResponse(response: response)
                    result("\(theJSONText ?? "")")
                }
                }
            }else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#revgeocode":
            let reverseGeocodeManager = MapmyIndiaReverseGeocodeManager.shared
            if let option = RestAPIHelper.reverseGeocodeOption(argument: call.arguments) {
                reverseGeocodeManager.reverseGeocode(option) { (placemarks, attribution, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    } else {
                        let theJSONText = RestAPIHelper.reverseGeocodeResponse(response: placemarks ?? [])
                        result("\(theJSONText ?? "")")
                    }
                }
            }else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#nearby":
            let nearByManager = MapmyIndiaNearByManager.shared
            if let option = RestAPIHelper.nearbyOption(argument: call.arguments) {
            nearByManager.getNearBySuggestions(option) { (nearbyResult, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    }else {
                        let theJSONText = RestAPIHelper.nearbyResponse(response: nearbyResult)
                        result("\(theJSONText ?? "")")
                    }
            }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#direction":
            let directionManager = Directions(restKey: MapmyIndiaAccountManager.restAPIKey())
            if let option = RestAPIHelper.directionOption(argument: call.arguments) {
                directionManager.calculate(option) { (waypoints, routes, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    }else {
                        let theJSONText = RestAPIHelper.directionResponse(response: routes ?? [], waypoints: waypoints ?? [], geometry: option.shapeFormat)
                        result("\(theJSONText ?? "")")
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#distancematrix":
            let distanceManager = MapmyIndiaDrivingDistanceMatrixManager.shared
            if let option = RestAPIHelper.distanceMatrix(argument: call.arguments) {
            distanceManager.getResult(option) { (distanceResult, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    }else {
                        //{"responseCode":200,"version":"202104.19.5223","results":{"code":"Ok","distances":[[137993.4]],"durations":[[10964.7]]}}
                        let theJSONText = RestAPIHelper.distanceMatrixResponse(response: distanceResult)
                        result("\(theJSONText ?? "")")
                    }
            }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#feedback":
            let feedbackManager = MapmyIndiaFeedbackManager.shared
            if let option = RestAPIHelper.feedbackOption(argument: call.arguments) {
                feedbackManager.sendFeedback(option) { (created, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    }else {
                        result(201)
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#poialongroute":
            let poiAlongRouteManager = MapmyIndiaPOIAlongTheRouteManager.shared
            if let option = RestAPIHelper.poiAlongRouteOption(argument: call.arguments) {
                poiAlongRouteManager.getPOIsAlongTheRoute(option) { (response, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    }else {
                        let theJSONText = RestAPIHelper.poiAlongRouteResponse(response: response)
                        result("\(theJSONText ?? "")")
                    }
                    
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#textsearch":
            let textSearchManager = MapmyIndiaAtlasTextSearchManager.shared
            if let option = RestAPIHelper.textSearchOption(argument: call.arguments) {
                textSearchManager.getTextSearchResult(option) { (suggestions, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    } else {
                        let theJSONText = RestAPIHelper.autosuggestResponse(response: suggestions)
                        result("\(theJSONText ?? "")")
                    }
                    
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#placedetail":
            let placeDetailManager = MapmyIndiaPlaceDetailManager.shared
            if let option = RestAPIHelper.placeDetailOption(argument: call.arguments) {
                placeDetailManager.getResults(option) { (placeDetail, error) in
                        if let error = error {
                            result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                        } else {
                            let theJSONText = RestAPIHelper.placeDetailResponse(response: placeDetail)
                            result("\(theJSONText ?? "")")
                        }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#placedetaillegacy":
            let placeDetailLegacyManager = MapmyIndiaPlaceDetailLegacyManager.shared
            if let option = RestAPIHelper.placeDetailLegacyOption(argument: call.arguments) {
                placeDetailLegacyManager.getPlaceDetail(option) { (placeDetail, attribution, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    } else {
                        let theJSONText = RestAPIHelper.placeDetailLegacyResponse(response: placeDetail)
                        result("\(theJSONText ?? "")")
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#hateosnearby":
            let hateOsManager = MapmyIndiaHateOSManager.shared
            if let option = RestAPIHelper.hateOsNearbyOption(argument: call.arguments) {
                hateOsManager.getResults(option) { (nearbyResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedFailureReason))
                    } else {
                        let theJSONText = RestAPIHelper.hateOsNearbyResponse(response: nearbyResponse)
                        result("\(theJSONText ?? "")")
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#geoanalyticslist":
            let listingManager = MapmyIndiaGeoanalyticsListingManager.shared()
            if let option = RestAPIHelper.geoAnalyticsListOption(argument: call.arguments) {
                listingManager.getListingInfo(option) { (nearbyResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(8), message: error.localizedDescription, details: error.localizedDescription))
                    } else {
                        let theJSONText = RestAPIHelper.geoAnalyticsListResponse(response: nearbyResponse)
                        result("\(theJSONText ?? "")")
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#generateotp":
            let generateOtpManager = MapmyIndiaOtpGenerateManager.shared
            if let option = RestAPIHelper.generateOtpOtion(argument: call.arguments) {
                generateOtpManager.getOtp(option) { (generateOtpResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedDescription))
                    } else {
                        result(generateOtpResponse)
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        case "rest#whitelist":
            let whitelistManager = MapmyIndiaWhitelistManager.shared
            if let option = RestAPIHelper.whitelist(argument: call.arguments) {
                whitelistManager.whitelistUser(option) {(whitelistResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedDescription))
                    } else {
                        result(nil)
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
        case "rest#roadtrafficdetail":
            let roadTrafficDetailManager = MapmyIndiaRoadTrafficDetailsManager.shared
            if let option = RestAPIHelper.roadTrafficDetail(argument: call.arguments) {
                roadTrafficDetailManager.getTrafficRoadDetailsResults(option){(trafficDetailResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedDescription))
                    } else {
                        let theJSONText = RestAPIHelper.trafficDetailResponse(response: trafficDetailResponse)
                        result("\(theJSONText ?? "")")
                    }
                }
            } else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
        case "rest#nearbyreport":
            let nearbyReportManager = MapmyIndiaNearbyReportManager.shared
            if let option = RestAPIHelper.nearbyReport(argument: call.arguments) {
                nearbyReportManager.getNearbyReportResult(option){(nearbyReportResponse, error) in
                    if let error = error {
                        result(FlutterError.init(code: String(error.code), message: error.localizedDescription, details: error.localizedDescription))
                    } else {
                        let theJSONText = RestAPIHelper.nearbyReportResponse(response: nearbyReportResponse)
                        result("\(theJSONText ?? "")")
                    }
                }
            }else {
                result(FlutterError(code: "-1", message: "Something went wrong", details: nil))
            }
            
        default:
            result(FlutterMethodNotImplemented)
            
        }
    }
    
}
