import Flutter
import UIKit
import MapmyIndiaMaps
import MapmyIndiaAPIKit

public class SwiftMapboxGlFlutterPlugin: NSObject, FlutterPlugin {
    public static func register(with registrar: FlutterPluginRegistrar) {
        let instance = MapboxMapFactory(withRegistrar: registrar)
        registrar.register(instance, withId: "plugins.flutter.io/mapmyindia_gl")
        
        let channel = FlutterMethodChannel(
            name: "plugins.flutter.io/mapmyindia_gl",
            binaryMessenger: registrar.messenger()
        )
        
        channel.setMethodCallHandler { methodCall, result in
            switch methodCall.method {
            case "installOfflineMapTiles":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let tilesdb = arguments["tilesdb"]
                installOfflineMapTiles(registrar: registrar, tilesdb: tilesdb!)
                result(nil)
            case "downloadOfflineRegion":
                // Get download region arguments from caller
                guard let args = methodCall.arguments as? [String: Any],
                      let definitionDictionary = args["definition"] as? [String: Any],
                      let metadata = args["metadata"] as? [String: Any],
                      let defintion = OfflineRegionDefinition.fromDictionary(definitionDictionary),
                      let channelName = args["channelName"] as? String
                else {
                    print(
                        "downloadOfflineRegion unexpected arguments: \(String(describing: methodCall.arguments))"
                    )
                    result(nil)
                    return
                }
                // Prepare channel
                let channelHandler = OfflineChannelHandler(
                    messenger: registrar.messenger(),
                    channelName: channelName
                )
                OfflineManagerUtils.downloadRegion(
                    definition: defintion,
                    metadata: metadata,
                    result: result,
                    registrar: registrar,
                    channelHandler: channelHandler
                )
            case "setOfflineTileCountLimit":
                guard let arguments = methodCall.arguments as? [String: Any],
                      let limit = arguments["limit"] as? UInt64
                else {
                    result(FlutterError(
                        code: "SetOfflineTileCountLimitError",
                        message: "could not decode arguments",
                        details: nil
                    ))
                    return
                }
                OfflineManagerUtils.setOfflineTileCountLimit(result: result, maximumCount: limit)
            case "getListOfRegions":
                // Note: this does not download anything from internet, it only fetches data drom database
                OfflineManagerUtils.regionsList(result: result)
            case "deleteOfflineRegion":
                guard let args = methodCall.arguments as? [String: Any],
                      let id = args["id"] as? Int
                else {
                    result(nil)
                    return
                }
                OfflineManagerUtils.deleteRegion(result: result, id: id)
            case "setRestAPIKey":
                guard let argumentsRestAPIKey = methodCall.arguments as? [String: String] else { return }
                let restAPIKey = argumentsRestAPIKey["restAPIKey"]
                MapmyIndiaAccountManager.setRestAPIKey(restAPIKey)
                result(nil)
                
            case "setMapSDKKey":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let mapSDKKey = arguments["mapSDKKey"]
                MapmyIndiaAccountManager.setMapSDKKey(mapSDKKey)
                result(nil)
                
            case "setAtlasClientId":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let atlasClientId = arguments["atlasClientId"]
                MapmyIndiaAccountManager.setAtlasClientId(atlasClientId)
                result(nil)
                
            case "setAtlasClientSecret":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let atlasClientSecret = arguments["atlasClientSecret"]
                MapmyIndiaAccountManager.setAtlasClientSecret(atlasClientSecret)
                result(nil)
                
            case "setAtlasGrantType":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let atlasGrantType = arguments["atlasGrantType"]
                MapmyIndiaAccountManager.setAtlasGrantType(atlasGrantType)
                result(nil)
                
            case "setShowLastSelectedStyle":
                guard let arguments = methodCall.arguments as? [String: Bool] else { return }
                let showLastSelectedStyle = arguments["showLastSelectedStyle"]
                MapmyIndiaMapConfiguration.setIsShowPreferedMapStyle(showLastSelectedStyle ?? true)
                result(nil)
                
            case "isShowLastSelectedStyle":
                let isShowLastSelectedStyle = MapmyIndiaMapConfiguration.isShowPreferedMapStyle()
                result(isShowLastSelectedStyle)
                
            case "getRestAPIKey":
                let restAPIKey = MapmyIndiaAccountManager.restAPIKey()
                result(restAPIKey)
                
            case "getAtlasClientId":
                let atlasClientId = MapmyIndiaAccountManager.atlasClientId()
                result(atlasClientId)
                
            case "getAtlasClientSecret":
                let atlasClientSecret = MapmyIndiaAccountManager.atlasClientSecret()
                result(atlasClientSecret)
                
            case "getAtlasGrantType":
                let atlasGrantType = MapmyIndiaAccountManager.atlasGrantType()
                result(atlasGrantType)
                
            case "getMapSDKKey":
                let mapSDKKey = MapmyIndiaAccountManager.mapSDKKey()
                result(mapSDKKey)
                
            case "setDeveloperShowingSplash":
                guard let arguments = methodCall.arguments as? [String: Bool] else { return }
                let developerShowingSplash = arguments["developerShowingSplash"]
                MapmyIndiaAccountManager.setIsHiddenAuthorizationFailOverlay(developerShowingSplash ?? false)
                result(nil)
                
            case "isDeveloperShowingSplash":
                let isDeveloperShowingSplash = MapmyIndiaAccountManager.isHiddenAuthorizationFailOverlay()
                result(isDeveloperShowingSplash)
                
                
            case "setClusterId":
                guard let arguments = methodCall.arguments as? [String: String] else { return }
                let clusterId = arguments["clusterId"]
                let deviceAlias = arguments["deviceAlias"]
                if(deviceAlias == nil) {
                    MapmyIndiaAccountManager.setClusterId(clusterId)
                } else {
                    MapmyIndiaAccountManager.setClusterId(clusterId! ,linkedDeviceAlias: deviceAlias!)
                }
                result(nil)
            case "getClusterId":
                let clusterId = MapmyIndiaAccountManager.clusterId()
                result(clusterId)
            case "getDeviceAlias":
                let deviceAlias = MapmyIndiaAccountManager.linkedDeviceAlias()
                result(deviceAlias)
            case "setRegion":
                guard let arguments = methodCall.arguments as? [String: String] else {return}
                let region = arguments["region"]
                if(region != nil) {
                MapmyIndiaAccountManager.setDefaultRegion(MMIRegionTypeIdentifier.init(rawValue: region!))
                } else {
                    MapmyIndiaAccountManager.setDefaultRegion(.india)
                }
                result(nil)
            case "getRegion":
                let region = MapmyIndiaAccountManager.defaultRegion().rawValue
                result(region)
            case "setAssociationId":
                guard let arguments = methodCall.arguments as? [String: String] else {return}
                let associationId = arguments["associationId"]
                MapmyIndiaAccountManager.setAssociationId(associationId)
                result(nil)
            case "getAssociationId":
                let associationId = MapmyIndiaAccountManager.associationId()
                result(associationId)
            default:
                result(FlutterMethodNotImplemented)
                
                
            }
            let restApiChannel = FlutterMethodChannel(name: "plugins.flutter.io/mapmyindia_rest_api", binaryMessenger: registrar.messenger())
            restApiChannel.setMethodCallHandler{(call, result) in
                RestAPIMethodCall.init(call: call, result: result)
            }
        }
    }
    
    private static func getTilesUrl() -> URL {
        guard var cachesUrl = FileManager.default.urls(
            for: .applicationSupportDirectory,
               in: .userDomainMask
        ).first,
              let bundleId = Bundle.main
                .object(forInfoDictionaryKey: kCFBundleIdentifierKey as String) as? String
        else {
            fatalError("Could not get map tiles directory")
        }
        cachesUrl.appendPathComponent(bundleId)
        cachesUrl.appendPathComponent(".mapbox")
        cachesUrl.appendPathComponent("cache.db")
        return cachesUrl
    }
    
    // Copies the "offline" tiles to where Mapbox expects them
    private static func installOfflineMapTiles(registrar: FlutterPluginRegistrar, tilesdb: String) {
        var tilesUrl = getTilesUrl()
        let bundlePath = getTilesDbPath(registrar: registrar, tilesdb: tilesdb)
        NSLog(
            "Cached tiles not found, copying from bundle... \(String(describing: bundlePath)) ==> \(tilesUrl)"
        )
        do {
            let parentDir = tilesUrl.deletingLastPathComponent()
            try FileManager.default.createDirectory(
                at: parentDir,
                withIntermediateDirectories: true,
                attributes: nil
            )
            if FileManager.default.fileExists(atPath: tilesUrl.path) {
                try FileManager.default.removeItem(atPath: tilesUrl.path)
            }
            try FileManager.default.copyItem(atPath: bundlePath!, toPath: tilesUrl.path)
            var resourceValues = URLResourceValues()
            resourceValues.isExcludedFromBackup = true
            try tilesUrl.setResourceValues(resourceValues)
        } catch {
            NSLog("Error copying bundled tiles: \(error)")
        }
    }
    
    private static func getTilesDbPath(registrar: FlutterPluginRegistrar,
                                       tilesdb: String) -> String?
    {
        if tilesdb.starts(with: "/") {
            return tilesdb
        } else {
            let key = registrar.lookupKey(forAsset: tilesdb)
            return Bundle.main.path(forResource: key, ofType: nil)
        }
    }
}
